from typing import Optional

from cosmpy.aerial.config import NetworkConfig


def get_network_config(name: str) -> Optional[NetworkConfig]:
    if name == 'fetchai-testnet':
        return NetworkConfig.latest_stable_testnet()
    return None



import argparse
import os

from jenesis.config import Config
from jenesis.contracts.deploy import deploy_contracts


def run(args: argparse.Namespace):
    cfg = Config.load(os.getcwd())

    if args.profile not in cfg.profiles:
        print(f'Invalid profile name. Expected one of {",".join(cfg.profiles.keys())}')
        return 1

    deploy_contracts(cfg, args.profile, os.getcwd(), args.key)
    return 0


def add_deploy_command(parser):
    deploy_cmd = parser.add_parser("deploy")
    deploy_cmd.add_argument(
        "-p", "--profile", default="testing", help="The profile to deploy"
    )
    deploy_cmd.add_argument("key", nargs='?', metavar="KEY", help="Deployer Key for all contracts")
    deploy_cmd.set_defaults(handler=run)



import argparse
import os
from tempfile import mkdtemp
import shutil
import subprocess
from jenesis.config import Config

TEMPLATE_GIT_URL = 'git@github.com:fetchai/jenesis-templates.git'


def run_add_contract(args: argparse.Namespace):
    template = args.template
    name = args.name
    branch = args.branch

    project_root = os.path.abspath(os.getcwd())
    contract_root = os.path.join(project_root, 'contracts', name)

    # check that we are actually running the command from the project root
    if not os.path.exists(os.path.join(project_root, 'jenesis.toml')):
        print('Please run command from project root')
        return False

    # check to see if the contract already exists
    if os.path.exists(contract_root):
        print(f'Contract "{name}" already exists')
        return False

    # create the temporary clone folder
    temp_clone_path = mkdtemp(prefix='jenesis-', suffix='-tmpl')

    # clone the templates folder out in the temporary file
    print('Downloading template...')
    cmd = ['git', 'clone', '--single-branch']
    if branch is not None:
        cmd += ['--branch', branch]
    cmd += [TEMPLATE_GIT_URL, '.']
    with open(os.devnull, 'w', encoding='utf8') as null_file:
        subprocess.check_call(cmd, stdout=null_file, stderr=subprocess.STDOUT, cwd=temp_clone_path)

    # find the target contract
    contract_template_path = os.path.join(temp_clone_path, 'contracts', template)
    if not os.path.isdir(contract_template_path):
        print(f'Unknown template {template}')
        return False
    print('Downloading template...complete')

    # process all the files as part of the template
    print('Rendering template...')
    for root, _, files in os.walk(contract_template_path):
        for filename in files:
            file_path = os.path.join(root, filename)
            rel_path = os.path.relpath(file_path, contract_template_path)

            output_filepath = os.path.join(contract_root, rel_path)
            os.makedirs(os.path.dirname(output_filepath), exist_ok=True)
            with open(file_path, 'r', encoding='utf8') as input_file:
                with open(output_filepath, 'w', encoding='utf8') as output_file:
                    contents = input_file.read()

                    # replace the templating parameters here
                    contents = contents.replace('<<NAME>>', name)

                    output_file.write(contents)
    print('Rendering template...complete')

    # clean up the temporary folder
    shutil.rmtree(temp_clone_path)

    Config.create_project(os.getcwd())

    return True



import argparse

from jenesis.cmd.add.contract import run_add_contract


def add_add_command(parser):
    add_cmd = parser.add_parser('add')
    subparsers = add_cmd.add_subparsers()

    add_contract_cmd = subparsers.add_parser('contract', help='Contract service templates')
    add_contract_cmd.add_argument('template', help='The name of the template to use')
    add_contract_cmd.add_argument('name', help='The name of contract to create')
    add_contract_cmd.add_argument('-b', '--branch', help='The name of the branch that should be used')
    add_contract_cmd.set_defaults(handler=run_add_contract)



import argparse

from jenesis.cmd.alpha.deploy import add_deploy_command
from jenesis.cmd.alpha.keys import add_keys_command
from jenesis.cmd.alpha.run import add_run_command
from jenesis.cmd.alpha.shell import add_shell_command


def add_alpha_command(parser):
    alpha_cmd = parser.add_parser('alpha')
    subparsers = alpha_cmd.add_subparsers()

    add_deploy_command(subparsers)
    add_keys_command(subparsers)
    add_run_command(subparsers)
    add_shell_command(subparsers)



import hashlib
import json
import os
import struct
import subprocess
import sys
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

import toml
from cosmpy.crypto.address import Address
from jenesis.config.errors import ConfigurationError
from jenesis.config.extract import (extract_opt_dict, extract_opt_int,
                                 extract_opt_str, extract_req_str,
                                 extract_req_str_list)
from jenesis.contracts import Contract
from jenesis.contracts.detect import detect_contracts


@dataclass
class Deployment:
    contract: Contract # internal: the contract that is deployed
    network: str  # internal: the name of the network to deploy to
    deployer_key: str  # config: the name of the key to use for deployment
    init: Any  # config: init parameters for the contract
    checksum: Optional[str]  # lock: the checksum digest to detect configuration changes
    digest: Optional[str]  # lock: the contract of the deployed contract
    address: Optional[Address]  # lock: the address of the deployed contract
    code_id: Optional[int]  # lock: the code of the deployed contract

    def reset_metadata(self):
        self.checksum = None
        self.digest = None
        self.address = None
        self.code_id = None

    def compute_checksum(self) -> str:
        hasher = hashlib.sha256()

        def update(value: Optional[Any]):
            encoded_value = "" if value is None else str(value)

            hasher.update(struct.pack(">Q", len(encoded_value)))
            hasher.update(encoded_value.encode())

        canonical_init_args = json.dumps(self.init, sort_keys=True)

        update(self.network)
        update(self.deployer_key)
        update(canonical_init_args)
        update(self.digest)
        update(self.address)
        update(self.code_id)

        return hasher.hexdigest()

    def __repr__(self) -> str:
        return f'{self.contract.name}: {self.address}'

    def is_configuration_out_of_date(self) -> bool:
        return self.checksum != self.compute_checksum()

    def to_lockfile(self) -> Any:
        data = {
            "checksum": self.compute_checksum(),
        }

        if self.digest is not None:
            data["digest"] = str(self.digest)
        if self.address is not None:
            data["address"] = str(self.address)
        if self.code_id is not None:
            data["code_id"] = self.code_id

        return data


@dataclass
class Profile:
    name: str
    network: str
    contracts: Dict[str, Contract]
    deployments: Dict[str, Deployment]

    def to_lockfile(self) -> Any:
        return {
            name: deployment.to_lockfile() for name, deployment in self.deployments.items()
        }


@dataclass
class Config:
    project_name: str
    project_authors: List[str]
    profiles: Dict[str, Profile]

    def update_deployment(
        self,
        profile_name: str,
        contract_name: str,
        digest: str,
        code_id: int,
        address: Address,
    ):
        profile = self.profiles.get(profile_name)
        if profile is None:
            raise ConfigurationError(f"unable to lookup profile {profile_name}")

        contract = profile.contracts.get(contract_name)
        deployment = profile.deployments.get(contract_name)
        if deployment is None:
            deployment = Deployment(
                contract,
                profile.network,
                "", None, None, None, None, None
            )

        # update the contract if necessary
        if digest is not None:
            deployment.digest = str(digest)
        if code_id is not None:
            deployment.code_id = int(code_id)
        if address is not None:
            deployment.address = Address(address)

        profile.deployments[contract_name] = deployment
        self.profiles[profile_name] = profile

    @classmethod
    def load(cls, path: str) -> "Config":
        project_file_path = os.path.join(path, "jenesis.toml")
        lock_file_path = os.path.join(path, "jenesis.lock")

        if not os.path.isfile(project_file_path):
            raise ConfigurationError('Missing project file: "jenesis.lock"')
        project_contents = toml.load(project_file_path)

        if os.path.isfile(lock_file_path):
            lock_file_contents = toml.load(lock_file_path)
        else:
            lock_file_contents = {}
            print("No lock file found. Proceeding without any contract deployments.")

        return cls._loads(project_contents, lock_file_contents)

    @classmethod
    def _loads(cls, project_contents: Any, lock_file_contents: Any) -> "Config":

        # extract the profiles section
        config_profiles = project_contents.get("profile", {})
        if not isinstance(config_profiles, dict):
            raise ConfigurationError("invalid profile configuration section")

        lock_profiles = lock_file_contents.get("profile", {})
        if not isinstance(lock_profiles, dict):
            raise ConfigurationError("invalid profile lock section")

        profiles = {}
        for name, config_item in config_profiles.items():
            lock_item = lock_profiles.get(name, {})

            profile = cls._parse_profile(name, config_item, lock_item)
            profiles[profile.name] = profile

        return Config(
            project_name=extract_req_str(project_contents, "project.name"),
            project_authors=extract_req_str_list(project_contents, "project.authors"),
            profiles=profiles,
        )

    @classmethod
    def _parse_profile(cls, name: str, profile: Any, lock_profile: Any) -> Profile:
        if not isinstance(profile, dict):
            raise ConfigurationError(
                "profile configuration invalid, expected dictionary"
            )
        if not isinstance(lock_profile, dict):
            raise ConfigurationError(
                "profile lock configuration invalid, expected dictionary"
            )

        network = extract_req_str(profile, "network")

        profile_contracts = profile.get("contracts", {})
        if not isinstance(profile_contracts, dict):
            raise ConfigurationError("invalid contracts section in config")

        deployments = {}
        if "contracts" in profile:
            for contract_name, contract_settings in profile_contracts.items():
                deployment_lock = lock_profile.get(contract_name, {})

                deployment = cls._parse_contract_config(
                    contract_settings, network, deployment_lock
                )
                deployments[contract_name] = deployment

        return Profile(
            name=str(name),
            network=network,
            contracts=profile_contracts,
            deployments=deployments,
        )

    @classmethod
    def _parse_contract_config(
        cls, contract: dict, network: str, lock: Any
    ) -> Deployment:
        if not isinstance(contract, dict):
            raise ConfigurationError(
                "contract configuration invalid, expected dictionary"
            )
        if not isinstance(lock, dict):
            raise ConfigurationError(
                "contract lock configuration invalid, expected dictionary"
            )

        def opt_address(value: Optional[str]) -> Optional[Address]:
            return None if value is None else Address(value)

        return Deployment(
            contract=contract,
            network=str(network),
            init=extract_opt_dict(contract, "init"),
            deployer_key=extract_req_str(contract, "deployer_key"),
            digest=extract_opt_str(lock, "digest"),
            address=opt_address(extract_opt_str(lock, "address")),
            code_id=extract_opt_int(lock, "code_id"),
            checksum=extract_opt_str(lock, "checksum"),
        )

    def save(self, path: str):
        contents = {
            "profile": {
                name: profile.to_lockfile() for name, profile in self.profiles.items()
            }
        }

        lock_file_path = os.path.join(path, "jenesis.lock")
        with open(lock_file_path, "w", encoding="utf-8") as lock_file:
            toml.dump(contents, lock_file)

    @staticmethod
    def create_project(path: str):
        user_name = subprocess.getoutput("git config user.name")
        user_email = subprocess.getoutput("git config user.email")
        authors = [f"{user_name} <{user_email}>"]

        # take the project name directly from the base name of the project
        project_root = os.path.abspath(path)
        project_name = os.path.basename(project_root)

        # detect contract source code and add placeholders for key contract data
        contracts = detect_contracts(project_root) or []

        contract_cfgs = {contract.name: Deployment(contract,
            "fetchai-testnet", "", {arg: "" for arg in contract.init_args()},
            None, None, None, None,
        ) for contract in contracts}

        profiles = {
            "testing": {
                "network": "fetchai-testnet",
                "contracts": {name: vars(cfg) for (name, cfg) in contract_cfgs.items()},
            }
        }

        data = {
            "project": {"name": project_name, "authors": authors},
            "profile": profiles,
        }

        project_configuration_file = os.path.join(project_root, "jenesis.toml")
        project_git_keep_files = [
            os.path.join(project_root, "contracts", ".gitkeep"),
        ]

        # create the configuration file
        os.makedirs(os.path.dirname(project_configuration_file), exist_ok=True)
        with open(project_configuration_file, "w", encoding="utf-8") as toml_file:
            toml.dump(data, toml_file)

        # create all the git kep files
        for git_keep_path in project_git_keep_files:
            os.makedirs(os.path.dirname(git_keep_path), exist_ok=True)

            with open(git_keep_path, "a", encoding="utf-8"):
                pass



import argparse
import os

from ptpython import embed

from cosmpy.aerial.client import LedgerClient

from jenesis.config import Config
from jenesis.contracts.detect import detect_contracts
from jenesis.contracts.monkey import MonkeyContract
from jenesis.contracts.networks import get_network_config
from jenesis.contracts.observer import DeploymentUpdater


def load_config(args: argparse.Namespace) -> dict:
    project_path = os.getcwd()

    cfg = Config.load(project_path)
    contracts = detect_contracts(project_path)

    print('Detecting contracts...')

    contract_instances = {}
    selected_profile = cfg.profiles.get(args.profile)
    if selected_profile is not None:
        net_config = get_network_config(selected_profile.network)
        if net_config is None:
            raise RuntimeError(f'Unknown network name "{selected_profile.network}"')

        # build the ledger client
        client = LedgerClient(net_config)

        for contract in contracts:
            print('C', contract)

            # skip contracts that we have not compiled
            if contract.digest() is None:
                continue

            # select the metadata for the contract
            selected_contract = selected_profile.deployments.get(contract.name)
            if selected_contract is not None:
                address = selected_contract.address
                code_id = selected_contract.code_id
            else:
                code_id = None
                address = None

            monkey = MonkeyContract(
                contract.binary_path,
                client,
                address=address,
                code_id=code_id,
                observer=DeploymentUpdater(
                    cfg,
                    project_path,
                    args.profile,
                    contract.name,
                ),
                init_args=selected_contract.init,
            )

            contract_instances[contract.name] = monkey

    print('Detecting contracts...complete')

    shell_globals = {}
    shell_globals['cfg'] = cfg
    shell_globals['project_path'] = project_path
    shell_globals['profile'] = args.profile
    shell_globals['contracts'] = {contract.name: contract for contract in contracts}
    for (name, instance) in contract_instances.items():
        shell_globals[name] = instance

    return shell_globals


def run(args: argparse.Namespace):

    shell_globals = load_config(args)
    shell_globals.update(globals())
    embed(shell_globals, vi_mode=False, history_filename='.shell_history')


def add_shell_command(parser):
    shell_cmd = parser.add_parser('shell')
    shell_cmd.add_argument('-p', '--profile', default='testing', help='The profile to use')
    shell_cmd.set_defaults(handler=run)



import argparse

from jenesis.config import Config


def run(args: argparse.Namespace):
    Config.create_project(args.project)


def add_new_command(parser):
    new_cmd = parser.add_parser("new")
    new_cmd.add_argument("project", metavar="NAME", help="Project name")
    new_cmd.set_defaults(handler=run)



import argparse
import os

from jenesis.config import Config


def run(_args: argparse.Namespace):
    Config.create_project(os.getcwd())


def add_init_command(parser):
    init_cmd = parser.add_parser("init")
    init_cmd.set_defaults(handler=run)



import os.path
from dataclasses import dataclass
from typing import Optional

from cosmpy.aerial.contract import _compute_digest


@dataclass
class Contract:
    name: str
    source_path: str
    binary_path: str
    schema: dict

    def digest(self) -> Optional[str]:
        if not os.path.isfile(self.binary_path):
            return None

        return _compute_digest(self.binary_path).hex()

    def messages(self) -> dict:
        return self.schema.keys()

    def init_args(self) -> dict:
        return self.schema['instantiate_msg']['properties'].keys()

    def __repr__(self):
        return self.name



import argparse

from .shell import load_config


def run(args: argparse.Namespace):
    shell_globals = load_config(args)
    with open(args.script_path, encoding='utf-8') as script:
        exec(script.read(), shell_globals)


def add_run_command(parser):
    run_cmd = parser.add_parser('run')
    run_cmd.add_argument('-p', '--profile', default='testing', help='The profile to use')
    run_cmd.add_argument('script_path', help='The path to the script to run')
    run_cmd.set_defaults(handler=run)



import argparse
import os

from blessings import Terminal

from jenesis.contracts.build import build_contracts, build_workspace
from jenesis.contracts.detect import detect_contracts, is_workspace


def run(args: argparse.Namespace):
    term = Terminal()

    contracts = detect_contracts(os.getcwd())
    if contracts is None or len(contracts) == 0:
        print(term.red('Unable to detect any contracts'))
        return 1

    if is_workspace(os.getcwd()):
        build_workspace(os.getcwd(), contracts)
        return 0

    # build the contracts
    build_contracts(contracts, batch_size=args.batch_size)
    return 0


def add_compile_command(parser):
    compile_cmd = parser.add_parser('compile')
    compile_cmd.add_argument('-p', dest='batch_size', type=int, help='The limit of the number of tasks to do in parallel')
    compile_cmd.set_defaults(handler=run)



import platform
from typing import Union, List

from jenesis.keyring.amino.codec import LocalInfo, OfflineInfo

try:
    from jenesis.keyring import macos
except Exception:
    pass


def query_keychain_item(name: str) -> Union[LocalInfo, OfflineInfo]:
    system = platform.system()
    if system == 'Darwin':
        return macos.query_keychain_item(name)

    raise RuntimeError(f'Platform {system} not currently supported')


def query_keychain_items() -> List[str]:
    system = platform.system()
    if system == 'Darwin':
        return macos.query_keychain_items()

    raise RuntimeError(f'Platform {system} not currently supported')
